#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <iterator>
#include <regex>
#include <cstdlib>

#include <TFile.h>
#include <TString.h>
#include <TTree.h>
#include <TStopwatch.h>

#include "mtEvent.h"
#include "mtCollision.h"
#include "mtCollisionHeader.h"
#include "mtNucleon.h"
#include "mtEventHeader.h"
#include "mtParticle.h"

int main(int argc, char **argv)
{
  TString inFileName, outFileName;
  Int_t verbosity = 0;
  if (argc < 5)
  {
    std::cerr << "./osc2mt -i INPUTFILE -o OUTPUTFILE [OPTIONAL: -v 0,1,2 (verbosity level)]" << std::endl;
    return 10;
  }
  for (int i = 1; i < argc; i++)
  {
    if (std::string(argv[i]) != "-i" &&
        std::string(argv[i]) != "-o" &&
        std::string(argv[i]) != "-v")
    {
      std::cerr << "\nosc2mt: Unknown parameter: " << argv[i] << std::endl;
      return 11;
    } else
    {
      if (std::string(argv[i]) == "-i" && i != argc - 1)
      {
        inFileName = argv[++i];
      }
      if (std::string(argv[i]) == "-i" && i == argc - 1)
      {
        std::cerr << "\nosc2mt: Input file name was not specified!" << std::endl;
        return 20;
      }
      if (std::string(argv[i]) == "-o" && i != argc - 1)
      {
        outFileName = argv[++i];
      }
      if (std::string(argv[i]) == "-o" && i == argc - 1)
      {
        std::cerr << "\nosc2mt: Output file name was not specified!" << std::endl;
        return 21;
      }
      if (std::string(argv[i]) == "-v" && i != argc - 1)
      {
        verbosity = std::strtol(argv[++i], nullptr, 0);
      }
      if (std::string(argv[i]) == "-v" && i == argc - 1)
      {
        std::cerr << "\nosc2mt: Verbosity!" << std::endl;
        return 20;
      }
    }
  }
  if (verbosity != 0 && verbosity != 1 && verbosity != 2)
  {
    std::cerr << "\nosc2mt: Incorrect verbosity level. Take it by default." << std::endl;
    verbosity = 0;
  }

  TStopwatch timer;
  timer.Start();

  std::ifstream dataFile;
  dataFile.open(inFileName.Data());

  if (!dataFile)
  {
    printf("File does not exist\n");
    return 1;
  } else
  {
    std::cout << Form("Successfully opened file %s", inFileName.Data()) << std::endl;
  }

  std::string linestr;
  std::vector<std::string> tokens;

  auto *fi = new TFile(outFileName.Data(), "recreate");
  std::cout << Form("Successfully created file %s", outFileName.Data()) << std::endl;
  std::cout << std::endl;

  auto *tree = new TTree("tree", "UrQMD OSCAR 1999A format");

  auto *event = new mtEvent();
  mtEventHeader *eventHeader;
  mtParticle *particle;
  mtCollision *collision;
  mtCollisionHeader *collHeader;
  mtNucleon *nucleonIn, *nucleonOut;

  //Event characterization variables
  int nspectator = 0;
  int numevent = 0;
  int nucl = 0;

  tree->Branch("mtEvent", &event, 32000, 99);

  while (dataFile)
  {
    tokens.clear();
    std::getline(dataFile, linestr);
    std::istringstream iss(linestr);
    std::copy(std::istream_iterator<std::string>(iss), std::istream_iterator<std::string>(),
              std::back_inserter(tokens));
    if (verbosity != 0) std::cout << "*     " << linestr << std::endl;

    //Ignore file header starting with #
    std::string st_input1, st_input2;
    if (tokens[0] == "#")
    {
      continue;
    }

    //Find new event header
    if (std::strtol(tokens[0].c_str(), nullptr,0) == 0 && std::strtol(tokens[1].c_str(), nullptr,0) != 0)
    {
      numevent++;

      if (verbosity == 0 && numevent % 1000 == 0) std::cout << "Event " << numevent << std::endl;

      nucl = (int)std::strtol(tokens[1].c_str(),nullptr,0);

      eventHeader = new mtEventHeader();
      eventHeader->SetId(numevent);
      eventHeader->SetB(std::strtod(tokens[3].c_str(), nullptr));
      eventHeader->SetPhi(std::strtod(tokens[4].c_str(), nullptr));

      event->SetEventHeader(eventHeader);
      delete eventHeader;

      collision = new mtCollision();

      collHeader = new mtCollisionHeader();
      collHeader->SetId(0);
      collHeader->SetNin(std::strtol(tokens[0].c_str(), nullptr,0));
      collHeader->SetNout(nucl);
      collHeader->SetProcessId(0);
      collision->SetCollisionHeader(collHeader);
      delete collHeader;

      //Read the initial nucleons before collisions
      for (int i = 0; i < nucl; i++)
      {
        tokens.clear();
        std::getline(dataFile, linestr);
        std::istringstream iss1(linestr);
        std::copy(std::istream_iterator<std::string>(iss1), std::istream_iterator<std::string>(),
                  std::back_inserter(tokens));
        if (verbosity != 0) std::cout << "**    " << linestr << std::endl;

        nucleonOut = new mtNucleon();
        nucleonOut->SetId(std::strtol(tokens[0].c_str(), nullptr,0));
        nucleonOut->SetPdg(std::strtol(tokens[1].c_str(), nullptr,0));
        nucleonOut->SetPx(std::strtod(tokens[3].c_str(), nullptr));
        nucleonOut->SetPy(std::strtod(tokens[4].c_str(), nullptr));
        nucleonOut->SetPz(std::strtod(tokens[5].c_str(), nullptr));
        nucleonOut->SetE(std::strtod(tokens[6].c_str(), nullptr));
        nucleonOut->SetM(std::strtod(tokens[7].c_str(), nullptr));
        nucleonOut->SetX(std::strtod(tokens[8].c_str(), nullptr));
        nucleonOut->SetY(std::strtod(tokens[9].c_str(), nullptr));
        nucleonOut->SetZ(std::strtod(tokens[10].c_str(), nullptr));
        nucleonOut->SetT(std::strtod(tokens[11].c_str(), nullptr));

        collision->AddFinalNucleon(nucleonOut);
        delete nucleonOut;
      }
      event->AddCollision(collision);
      delete collision;

      continue;
    }
    //Find collision header in the current event for 2 -> many scatterings
    if (tokens[0] == "2")
    {
      int numcollproducts = (int)std::strtol(tokens[1].c_str(), nullptr,0);

      collision = new mtCollision();

      collHeader = new mtCollisionHeader();
      collHeader->SetId(std::strtol(tokens[3].c_str(), nullptr,0));
      collHeader->SetNin(std::strtol(tokens[0].c_str(), nullptr,0));
      collHeader->SetNout(numcollproducts);
      collHeader->SetProcessId(std::strtol(tokens[2].c_str(), nullptr,0));
      collision->SetCollisionHeader(collHeader);
      delete collHeader;

      tokens.clear();
      std::getline(dataFile, linestr);
      std::istringstream iss2(linestr);
      std::copy(std::istream_iterator<std::string>(iss2), std::istream_iterator<std::string>(),
                std::back_inserter(tokens));
      if (verbosity != 0) std::cout << "+     " << linestr << std::endl;

      //Is this scattering between nucleons?
      if (std::strtol(tokens[0].c_str(), nullptr,0) >= 1 && std::strtol(tokens[0].c_str(), nullptr,0) <= nucl)
      {

        nucleonIn = new mtNucleon();
        nucleonIn->SetId(std::strtol(tokens[0].c_str(), nullptr,0));
        nucleonIn->SetPdg(std::strtol(tokens[1].c_str(), nullptr,0));
        nucleonIn->SetPx(std::strtod(tokens[3].c_str(), nullptr));
        nucleonIn->SetPy(std::strtod(tokens[4].c_str(), nullptr));
        nucleonIn->SetPz(std::strtod(tokens[5].c_str(), nullptr));
        nucleonIn->SetE(std::strtod(tokens[6].c_str(), nullptr));
        nucleonIn->SetM(std::strtod(tokens[7].c_str(), nullptr));
        nucleonIn->SetX(std::strtod(tokens[8].c_str(), nullptr));
        nucleonIn->SetY(std::strtod(tokens[9].c_str(), nullptr));
        nucleonIn->SetZ(std::strtod(tokens[10].c_str(), nullptr));
        nucleonIn->SetT(std::strtod(tokens[11].c_str(), nullptr));

        tokens.clear();
        std::getline(dataFile, linestr);
        std::istringstream iss3(linestr);
        std::copy(std::istream_iterator<std::string>(iss3), std::istream_iterator<std::string>(),
                  std::back_inserter(tokens));
        if (verbosity != 0) std::cout << "+     " << linestr << std::endl;

        if (std::strtol(tokens[0].c_str(), nullptr,0) >= 1 && std::strtol(tokens[0].c_str(), nullptr,0) <= nucl)
        {
          nucleonOut = new mtNucleon();
          nucleonOut->SetId(std::strtol(tokens[0].c_str(), nullptr,0));
          nucleonOut->SetPdg(std::strtol(tokens[1].c_str(), nullptr,0));
          nucleonOut->SetPx(std::strtod(tokens[3].c_str(), nullptr));
          nucleonOut->SetPy(std::strtod(tokens[4].c_str(), nullptr));
          nucleonOut->SetPz(std::strtod(tokens[5].c_str(), nullptr));
          nucleonOut->SetE(std::strtod(tokens[6].c_str(), nullptr));
          nucleonOut->SetM(std::strtod(tokens[7].c_str(), nullptr));
          nucleonOut->SetX(std::strtod(tokens[8].c_str(), nullptr));
          nucleonOut->SetY(std::strtod(tokens[9].c_str(), nullptr));
          nucleonOut->SetZ(std::strtod(tokens[10].c_str(), nullptr));
          nucleonOut->SetT(std::strtod(tokens[11].c_str(), nullptr));

          collision->AddInitialNucleon(nucleonIn);
          collision->AddInitialNucleon(nucleonOut);
          delete nucleonOut;

          // Read FinalNucleons
          for (int i = 0; i < numcollproducts; i++)
          {
            tokens.clear();
            std::getline(dataFile, linestr);
            std::istringstream iss4(linestr);
            std::copy(std::istream_iterator<std::string>(iss4), std::istream_iterator<std::string>(),
                      std::back_inserter(tokens));
            if (verbosity != 0) std::cout << "++    " << linestr << std::endl;

            nucleonOut = new mtNucleon();
            nucleonOut->SetId(std::strtol(tokens[0].c_str(), nullptr,0));
            nucleonOut->SetPdg(std::strtol(tokens[1].c_str(), nullptr,0));
            nucleonOut->SetPx(std::strtod(tokens[3].c_str(), nullptr));
            nucleonOut->SetPy(std::strtod(tokens[4].c_str(), nullptr));
            nucleonOut->SetPz(std::strtod(tokens[5].c_str(), nullptr));
            nucleonOut->SetE(std::strtod(tokens[6].c_str(), nullptr));
            nucleonOut->SetM(std::strtod(tokens[7].c_str(), nullptr));
            nucleonOut->SetX(std::strtod(tokens[8].c_str(), nullptr));
            nucleonOut->SetY(std::strtod(tokens[9].c_str(), nullptr));
            nucleonOut->SetZ(std::strtod(tokens[10].c_str(), nullptr));
            nucleonOut->SetT(std::strtod(tokens[11].c_str(), nullptr));

            collision->AddFinalNucleon(nucleonOut);
            delete nucleonOut;
          }
          delete nucleonIn;
          event->AddCollision(collision);
          delete collision;
          continue;
        }
        delete nucleonIn;

        for (int i = 0; i < numcollproducts; i++)
        {
          std::getline(dataFile, linestr);
          if (verbosity != 0) std::cout << "***   " << linestr << std::endl;
        }

        delete collision;
        continue;
      }

      //Skip over the rest of the lines corresponding to collision products
      for (int i = 0; i < numcollproducts + 1; i++)
      {
        std::getline(dataFile, linestr);
        if (verbosity != 0) std::cout << "***   " << linestr << std::endl;
      }
      continue;
    }
    //Identify final-state particles and store them
    if (std::strtol(tokens[0].c_str(), nullptr,0) > 0 && tokens[1] == "0" && tokens.size() == 2)
    {
      if (verbosity != 0) std::cout << "End of the event block" << std::endl;
      // Reading particles
      for (int i = 0; i < (int)std::strtol(tokens[0].c_str(), nullptr,0); i++)
      {
        std::vector<std::string> finalstate_tokens;
        std::getline(dataFile, linestr);
        std::istringstream iss5(linestr);
        std::copy(std::istream_iterator<std::string>(iss5), std::istream_iterator<std::string>(),
                  std::back_inserter(finalstate_tokens));

        particle = new mtNucleon();
        particle->SetId(std::strtol(finalstate_tokens[0].c_str(), nullptr,0));
        particle->SetPdg(std::strtol(finalstate_tokens[1].c_str(), nullptr,0));
        particle->SetPx(std::strtod(finalstate_tokens[3].c_str(), nullptr));
        particle->SetPy(std::strtod(finalstate_tokens[4].c_str(), nullptr));
        particle->SetPz(std::strtod(finalstate_tokens[5].c_str(), nullptr));
        particle->SetE(std::strtod(finalstate_tokens[6].c_str(), nullptr));
        particle->SetM(std::strtod(finalstate_tokens[7].c_str(), nullptr));
        particle->SetX(std::strtod(finalstate_tokens[8].c_str(), nullptr));
        particle->SetY(std::strtod(finalstate_tokens[9].c_str(), nullptr));
        particle->SetZ(std::strtod(finalstate_tokens[10].c_str(), nullptr));
        particle->SetZ(std::strtod(finalstate_tokens[11].c_str(), nullptr));

        event->AddParticle(particle);
        delete particle;

        if (std::strtol(finalstate_tokens[0].c_str(), nullptr,0) <= nucl) nspectator++;
        if (verbosity != 0) std::cout << "****  " << linestr << std::endl;
      }
    }

    // tree->Fill();
    // event->Clear();
    // Finding end of this event
    if (tokens[0] == "0" && tokens[1] == "0" && tokens.size() == 2)
    {
      if (verbosity != 0) std::cout << "||||| Write event in tree.  |||||" << std::endl;
      tree->Fill();
      event->Clear();
      continue;
    }
    if (!dataFile) break;
  }

  fi->Write();
  tree->Print();

  std::cout << std::endl;

  timer.Stop();
  timer.Print();

  return 0;
}